/*
  "Timer2TaskScheduler"
  V1.0 / Copyright (C) 2017, T.Rode (DL1DUZ)

  Simple Task-scheduler library for Arduino using Timer 2
 (conflicting with the tone() and analogWrite() at pins 3 & 11).

  Permission is granted to use, copy, modify, and distribute this software
  and documentation for non-commercial purposes.

*/

//################################################################################################################
//includes
//################################################################################################################

#include <Timer2TaskScheduler.h>

//################################################################################################################
//declarations
//################################################################################################################

SchedulerClass* SchedulerClass::this_instance;

//################################################################################################################
//functions
//################################################################################################################

ISR(TIMER2_COMPA_vect) {
  SchedulerClass::this_instance->isr();
}

//################################################################################################################

// initializes the scheduler
// this function is to be called once; subsequent calls will have no effect
// returns true, if successful

// size -> maximum number of tasks the scheduler must be capable of handling (1...255)
// cycle -> time-basis (1=100μs; from 1...10 = 100μs...1ms)

boolean SchedulerClass::init(byte size, byte cycle) {
  this_instance = this;

  if(!is_initialized && size && cycle && cycle<11) {
    elements = size;
    // allocate memory for the task-array
    task_array_pointer = (task*) malloc(elements*sizeof(task));

    // initialize time 2
    cli(); // disable global interrupts
    TCCR2A = 0; // set entire TCCR2A register to 0
    TCCR2B = 0b00000100; // prescaler 64
    TCNT2  = 0; //initialize counter value to 0
    OCR2A = 25*cycle - 1; // set compare match register
    TCCR2A |= (1 << WGM21);  // turn on CTC mode
    TIMSK2 |= (1 << OCIE2A); // enable timer compare interrupt
    sei(); // enable global interrupts

    is_initialized = 1;
  }

  return is_initialized;
}

//################################################################################################################

// ISR of this instance
void SchedulerClass::isr() {
	
  static uint8_t i;
  static task* ptr;
 
  ptr = task_array_pointer;
  
  for(i=0; i<next_free_element; i++) {
    if(ptr->duetime) { ptr->duetime--; }
    ptr++;
  }

}

//################################################################################################################

// executes a task if due
void SchedulerClass::runIfDue(byte task) {
  dummy = task_array_pointer + task;

  ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
    if(task<next_free_element && dummy->is_scheduled && !(dummy->duetime)) {
      sei();
      dummy->is_scheduled = 0;
      dummy->f_pointer();
    }
  }
}

//################################################################################################################

// returns the max number of tasks the scheduler can handle
byte SchedulerClass::getTotalSize() {
  return elements;
}

//################################################################################################################

// returns the number of tasks that can still be added
byte SchedulerClass::getUsedSize() {
  return next_free_element;
}

//################################################################################################################

// returns the number of tasks that can still be added
byte SchedulerClass::getFreeSize() {
  return elements-next_free_element;
}

//################################################################################################################

// returns the due time in cycles or 0 if not scheduled or none-existend
unsigned long SchedulerClass::getDueTime(f_ptr function) {
  return processTask(function, 2, 0);
}

//################################################################################################################

// returns true if the task is scheduled or 0 if not scheduled or none-existend
boolean SchedulerClass::isScheduled(f_ptr function) {
  return processTask(function, 3, 0);
}

//################################################################################################################

// schedules a task to execute in x cycles /// the task is being registered if if does not yet exist
// cycles can be 0...2^32-1, representing 0s, 100μs ... 49.71026961805d (if cycle is set to 10 = 1ms)
// returns true if successful
boolean SchedulerClass::scheduleTask(f_ptr function, unsigned long time) {
  uint8_t target_scheduled = 1;

  if(!processTask(function, 1, time)) {
    if(getFreeSize()) {
      dummy = task_array_pointer + next_free_element;
      dummy->f_pointer = function;
      ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
        dummy->duetime = time;
      }
      dummy->is_scheduled = 1;

      next_free_element++;
    }
    else { target_scheduled = 0; }
  }

  return target_scheduled;
}

//################################################################################################################

// unschedules a task /// returns true if successful
boolean SchedulerClass::unscheduleTask(f_ptr function) {
  return processTask(function, 0, 0);
}

//################################################################################################################

// for internal use
// multi-purpouse: 0=unschedule task; 1=schedule task; 2=get due-time; 3=check if scheduled
// if task exists, 1 (in case of id=2 the due time in cycles) will be returned
uint32_t SchedulerClass::processTask(f_ptr function, uint8_t id, uint32_t time) {
  uint32_t target_found = 0;
  dummy = task_array_pointer;

  for(uint8_t i=0; i<next_free_element; i++) {
    if(dummy->f_pointer == function) {
      target_found = 1;

      ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
        switch(id) {
          case 0:
            dummy->duetime = 0;
            dummy->is_scheduled = 0;
          break;
          case 1:
            dummy->duetime = time;
            dummy->is_scheduled = 1;
          break;
          case 2:
            target_found = dummy->duetime;
          break;
          case 3:
            target_found = dummy->is_scheduled;
          break;
        }
      }

      break;
    }

    dummy++;
  }

  return target_found;
}

//################################################################################################################

SchedulerClass SDL;
